@@ -30,6 +30,10 @@
 
 #include "disk_control_interface.h"
 
+#ifdef HAVE_GFX_WIDGETS
+#include "gfx/gfx_widgets.h"
+#endif
+
 #ifdef HAVE_CHEEVOS
 #include "cheevos/cheevos.h"
 #endif
@@ -243,9 +247,9 @@ unsigned disk_control_get_image_index(
  **/
 void disk_control_get_image_label(
       disk_control_interface_t *disk_control,
-      unsigned index, char *label, size_t len)
+      unsigned index, char *s, size_t len)
 {
-   if (!label || len < 1)
+   if (!s || len < 1)
       return;
 
    if (!disk_control)
@@ -254,13 +258,13 @@ void disk_control_get_image_label(
    if (!disk_control->cb.get_image_label)
       goto error;
 
-   if (!disk_control->cb.get_image_label(index, label, len))
+   if (!disk_control->cb.get_image_label(index, s, len))
       goto error;
 
    return;
 
 error:
-   label[0] = '\0';
+   s[0] = '\0';
 }
 
 /***********/
@@ -273,58 +277,55 @@ error:
  * Generates an appropriate log/notification message
  * for a disk index change event
  **/
-static void disk_control_get_index_set_msg(
+static size_t disk_control_get_index_set_msg(
       disk_control_interface_t *disk_control,
       unsigned num_images, unsigned index, bool success,
-      unsigned *msg_duration, char *msg, size_t len)
+      unsigned *msg_duration, char *s, size_t len)
 {
-   bool has_label = false;
-   char image_label[128];
-
-   image_label[0] = '\0';
-
-   if (!disk_control || !msg_duration || !msg || len < 1)
-      return;
-
-   /* Attempt to get image label */
-   if (index < num_images)
-   {
-      disk_control_get_image_label(
-            disk_control, index, image_label, sizeof(image_label));
-      has_label = !string_is_empty(image_label);
-   }
-
-   /* Get message duration
-    * > Default is 60
-    * > If a label is shown, then increase duration by 50%
-    * > For errors, duration is always 180 */
-   *msg_duration = success ?
-         (has_label ? 90 : 60) :
-         180;
-
+   size_t _len    = 0;
+   if (!disk_control || !msg_duration || !s || len < 1)
+      return 0;
    /* Check whether image was inserted or removed */
    if (index < num_images)
    {
-      size_t _len = strlcpy(msg,
+      char image_label[128];
+      bool has_label = false;
+      image_label[0] = '\0';
+      disk_control_get_image_label(
+            disk_control, index, image_label, sizeof(image_label));
+
+      has_label      = !string_is_empty(image_label);
+
+      /* Get message duration
+       * > Default is 60
+       * > If a label is shown, then increase duration by 50%
+       * > For errors, duration is always 180 */
+      *msg_duration  = success ? (has_label ? 90 : 60) : 180;
+
+      _len = strlcpy(s,
             success
             ? msg_hash_to_str(MSG_SETTING_DISK_IN_TRAY)
             : msg_hash_to_str(MSG_FAILED_TO_SET_DISK), len);
       if (has_label)
-         snprintf(
-               msg + _len, len - _len, ": %u/%u - %s",
+         _len += snprintf(
+               s + _len, len - _len, ": %u/%u - %s",
                index + 1, num_images, image_label);
       else
-         snprintf(
-               msg + _len, len - _len, ": %u/%u",
+         _len += snprintf(
+               s + _len, len - _len, ": %u/%u",
                index + 1, num_images);
    }
    else
-      strlcpy(
-            msg,
+   {
+      *msg_duration = success ? 60 : 180;
+      _len += strlcpy(
+            s,
             success
             ? msg_hash_to_str(MSG_REMOVED_DISK_FROM_TRAY)
             : msg_hash_to_str(MSG_FAILED_TO_REMOVE_DISK_FROM_TRAY),
             len);
+   }
+   return _len;
 }
 
 /**
@@ -336,7 +337,7 @@ bool disk_control_set_eject_state(
       disk_control_interface_t *disk_control,
       bool eject, bool verbosity)
 {
-   bool error = false;
+   bool err = false;
    char msg[128];
    size_t _len;
 
@@ -353,8 +354,8 @@ bool disk_control_set_eject_state(
               sizeof(msg));
    else
    {
-      error = true;
-      _len  = strlcpy(
+      err  = true;
+      _len = strlcpy(
             msg,
             eject
             ? msg_hash_to_str(MSG_VIRTUAL_DISK_TRAY_EJECT)
@@ -364,20 +365,21 @@ bool disk_control_set_eject_state(
 
    if (_len > 0)
    {
-      if (error)
-         RARCH_ERR("[Disc]: %s\n", msg);
+      if (err)
+         RARCH_ERR("[Disc] %s\n", msg);
       else
-         RARCH_LOG("[Disc]: %s\n", msg);
+         RARCH_LOG("[Disc] %s\n", msg);
 
       /* Errors should always be displayed */
-      if (verbosity || error)
+      if (verbosity || err)
          runloop_msg_queue_push(
-               msg, _len, 1, error ? 180 : 60, true, NULL,
-               MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_INFO);
+               msg, _len, 2, err ? 180 : 60, true, NULL,
+               MESSAGE_QUEUE_ICON_DEFAULT,
+               err ? MESSAGE_QUEUE_CATEGORY_ERROR : MESSAGE_QUEUE_CATEGORY_INFO);
    }
 
 #ifdef HAVE_CHEEVOS
-   if (!error && !eject)
+   if (!err && !eject)
    {
       if (disk_control->cb.get_image_index && disk_control->cb.get_image_path)
       {
@@ -390,7 +392,7 @@ bool disk_control_set_eject_state(
    }
 #endif
 
-   return !error;
+   return !err;
 }
 
 /**
@@ -404,7 +406,8 @@ bool disk_control_set_index(
       disk_control_interface_t *disk_control,
       unsigned index, bool verbosity)
 {
-   bool error            = false;
+   size_t _len;
+   bool err              = false;
    unsigned num_images   = 0;
    unsigned msg_duration = 0;
    char msg[128];
@@ -427,30 +430,37 @@ bool disk_control_set_index(
    num_images = disk_control->cb.get_num_images();
 
    /* Perform 'set index' action */
-   error = !disk_control->cb.set_image_index(index);
+   err = !disk_control->cb.set_image_index(index);
 
    /* Get log/notification message */
-   disk_control_get_index_set_msg(
-         disk_control, num_images, index, !error,
+   _len = disk_control_get_index_set_msg(
+         disk_control, num_images, index, !err,
          &msg_duration, msg, sizeof(msg));
 
    /* Output log/notification message */
-   if (!string_is_empty(msg))
+   if (_len > 0)
    {
-      if (error)
-         RARCH_ERR("[Disc]: %s\n", msg);
+      if (err)
+         RARCH_ERR("[Disc] %s\n", msg);
       else
-         RARCH_LOG("[Disc]: %s\n", msg);
+         RARCH_LOG("[Disc] %s\n", msg);
 
       /* Errors should always be displayed */
-      if (verbosity || error)
-         runloop_msg_queue_push(msg, strlen(msg), 1, msg_duration, true, NULL,
+      if (verbosity)
+#ifdef HAVE_GFX_WIDGETS
+         gfx_widget_set_generic_message(msg, msg_duration * 10);
+#else
+         runloop_msg_queue_push(msg, _len, 2, msg_duration, true, NULL,
                MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_INFO);
+#endif
+      else if (err)
+         runloop_msg_queue_push(msg, _len, 2, msg_duration, true, NULL,
+               MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_ERROR);
    }
 
    /* If operation was successful, update disk
     * index record (if enabled) */
-   if (!error && disk_control->record_enabled)
+   if (!err && disk_control->record_enabled)
    {
       if (   disk_control->cb.get_image_index
           && disk_control->cb.get_image_path)
@@ -471,7 +481,7 @@ bool disk_control_set_index(
       }
    }
 
-   return !error;
+   return !err;
 }
 
 /**
@@ -504,7 +514,7 @@ bool disk_control_set_index_next(
 
    if (!disk_next_enable)
    {
-      RARCH_ERR("[Disc]: %s\n", msg_hash_to_str(MSG_GOT_INVALID_DISK_INDEX));
+      RARCH_ERR("[Disc] %s\n", msg_hash_to_str(MSG_GOT_INVALID_DISK_INDEX));
       return false;
    }
 
@@ -543,7 +553,7 @@ bool disk_control_set_index_prev(
 
    if (!disk_prev_enable)
    {
-      RARCH_ERR("[Disc]: %s\n", msg_hash_to_str(MSG_GOT_INVALID_DISK_INDEX));
+      RARCH_ERR("[Disc] %s\n", msg_hash_to_str(MSG_GOT_INVALID_DISK_INDEX));
       return false;
    }
 
@@ -625,13 +635,13 @@ bool disk_control_append_image(
       goto error;
 
    /* Display log */
-   _len        = strlcpy(msg, msg_hash_to_str(MSG_APPENDED_DISK), sizeof(msg));
+   _len        = strlcpy(msg, msg_hash_to_str(MSG_APPENDED_DISK), sizeof(msg) - 3);
    msg[  _len] = ':';
    msg[++_len] = ' ';
    msg[++_len] = '\0';
    _len += strlcpy(msg + _len, image_filename, sizeof(msg) - _len);
 
-   RARCH_LOG("[Disc]: %s\n", msg);
+   RARCH_LOG("[Disc] %s\n", msg);
    /* This message should always be displayed, since
     * the menu itself does not provide sufficient
     * visual feedback */
@@ -655,14 +665,14 @@ error:
       disk_control_set_eject_state(disk_control, false, false);
 
    _len        = strlcpy(msg,
-         msg_hash_to_str(MSG_FAILED_TO_APPEND_DISK), sizeof(msg));
+         msg_hash_to_str(MSG_FAILED_TO_APPEND_DISK), sizeof(msg) - 3);
    msg[  _len] = ':';
    msg[++_len] = ' ';
    msg[++_len] = '\0';
    _len += strlcpy(msg + _len, image_filename, sizeof(msg) - _len);
 
-   runloop_msg_queue_push(msg, _len, 0, 180, true, NULL,
-         MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_INFO);
+   runloop_msg_queue_push(msg, _len, 2, 180, true, NULL,
+         MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_WARNING);
 
    return false;
 }
@@ -721,7 +731,7 @@ bool disk_control_set_initial_index(
           * here may not matter (have to wait until
           * disk index is verified) */
          RARCH_ERR(
-               "[Disc]: Failed to set initial disk index: [%u] %s\n",
+               "[Disc] Failed to set initial disk index: #%u \"%s\".\n",
                disk_control->index_record.image_index,
                disk_control->index_record.image_path);
          return false;
@@ -799,10 +809,13 @@ bool disk_control_verify_initial_index(
    /* If current disk is incorrect, notify user */
    if (!success && enabled)
    {
-      const char *_msg = msg_hash_to_str(MSG_FAILED_TO_SET_INITIAL_DISK);
+      char _msg[128];
+      size_t _len = strlcpy(_msg,
+            msg_hash_to_str(MSG_FAILED_TO_SET_INITIAL_DISK), sizeof(_msg));
+
       RARCH_ERR(
-               "[Disc]: Failed to set initial disc index:\n> Expected"
-               " [%u] %s\n> Detected [%u] %s\n",
+               "[Disc] Failed to set initial disc index. Expected"
+               " #%u \"%s\", Detected #%u \"%s\".\n",
                disk_control->index_record.image_index + 1,
                disk_control->index_record.image_path,
                image_index + 1,
@@ -810,8 +823,8 @@ bool disk_control_verify_initial_index(
 
       /* Ignore 'verbosity' setting - errors should
        * always be displayed */
-      runloop_msg_queue_push(_msg, strlen(_msg), 0, 60, true, NULL,
-            MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_INFO);
+      runloop_msg_queue_push(_msg, _len, 0, 60, true, NULL,
+            MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_ERROR);
 
       /* Since a failure here typically means that the
        * original M3U content file has been altered,
@@ -835,16 +848,17 @@ bool disk_control_verify_initial_index(
     * is available */
    if (disk_control->initial_num_images > 1)
    {
+      size_t _len;
       unsigned msg_duration = 0;
       char msg[128];
 
       msg[0] = '\0';
 
-      disk_control_get_index_set_msg(
+      _len = disk_control_get_index_set_msg(
             disk_control, disk_control->initial_num_images, image_index, true,
             &msg_duration, msg, sizeof(msg));
 
-      RARCH_LOG("[Disc]: %s\n", msg);
+      RARCH_LOG("[Disc] %s\n", msg);
 
       /* Note: Do not flush message queue here, since
        * it is likely other notifications will be
@@ -852,7 +866,7 @@ bool disk_control_verify_initial_index(
        * we do not want to 'overwrite' them */
       if (verbosity)
          runloop_msg_queue_push(
-               msg, strlen(msg), 0, msg_duration, false, NULL,
+               msg, _len, 2, msg_duration, false, NULL,
                MESSAGE_QUEUE_ICON_DEFAULT, MESSAGE_QUEUE_CATEGORY_INFO);
 
 #ifdef HAVE_CHEEVOS
