@@ -25,6 +25,11 @@
 #include <process.h>
 #endif
 
+#if defined(_WIN32_WINNT) && _WIN32_WINNT >= 0x0600
+#include <winreg.h>
+#include <winerror.h>
+#endif
+
 #include <boolean.h>
 #include <compat/strl.h>
 #include <dynamic/dylib.h>
@@ -45,19 +50,12 @@
 #include "../frontend_driver.h"
 #include "../../configuration.h"
 #include "../../defaults.h"
-#include "../../verbosity.h"
-#include "../../ui/drivers/ui_win32.h"
 #include "../../paths.h"
 #include "../../msg_hash.h"
-#include "platform_win32.h"
-
 #include "../../verbosity.h"
+#include "../../ui/drivers/ui_win32.h"
 
-/*
-#ifdef HAVE_NVDA
-#include "../../nvda_controller.h"
-#endif
-*/
+#include "platform_win32.h"
 
 #ifdef HAVE_SAPI
 #define COBJMACROS
@@ -80,7 +78,7 @@ enum platform_win32_flags
 };
 
 #ifdef HAVE_SAPI
-static ISpVoice* pVoice           = NULL;
+static ISpVoice *voice_ptr        = NULL;
 #endif
 #ifdef HAVE_NVDA
 static uint8_t g_plat_win32_flags = PLAT_WIN32_FLAG_USE_NVDA;
@@ -123,34 +121,34 @@ const struct win32_lang_pair win32_lang_pairs[] =
    /* array order MUST be kept, always largest ID first */
    {0x7c04, RETRO_LANGUAGE_CHINESE_TRADITIONAL}, /* neutral */
    {0x1404, RETRO_LANGUAGE_CHINESE_TRADITIONAL}, /* MO */
-   {0x1004, RETRO_LANGUAGE_CHINESE_SIMPLIFIED}, /* SG */
-   {0xC04, RETRO_LANGUAGE_CHINESE_TRADITIONAL}, /* HK/PRC */
-   {0x816, RETRO_LANGUAGE_PORTUGUESE_PORTUGAL},
-   {0x416, RETRO_LANGUAGE_PORTUGUESE_BRAZIL},
-   {0x2a, RETRO_LANGUAGE_VIETNAMESE},
-   {0x19, RETRO_LANGUAGE_RUSSIAN},
-   {0x16, RETRO_LANGUAGE_PORTUGUESE_PORTUGAL},
-   {0x15, RETRO_LANGUAGE_POLISH},
-   {0x13, RETRO_LANGUAGE_DUTCH},
-   {0x12, RETRO_LANGUAGE_KOREAN},
-   {0x11, RETRO_LANGUAGE_JAPANESE},
-   {0x10, RETRO_LANGUAGE_ITALIAN},
-   {0xc, RETRO_LANGUAGE_FRENCH},
-   {0xa, RETRO_LANGUAGE_SPANISH},
-   {0x9, RETRO_LANGUAGE_ENGLISH},
-   {0x8, RETRO_LANGUAGE_GREEK},
-   {0x7, RETRO_LANGUAGE_GERMAN},
-   {0x4, RETRO_LANGUAGE_CHINESE_SIMPLIFIED}, /* neutral */
-   {0x1, RETRO_LANGUAGE_ARABIC},
+   {0x1004, RETRO_LANGUAGE_CHINESE_SIMPLIFIED},  /* SG */
+   {0xC04,  RETRO_LANGUAGE_CHINESE_TRADITIONAL}, /* HK/PRC */
+   {0x816,  RETRO_LANGUAGE_PORTUGUESE_PORTUGAL},
+   {0x416,  RETRO_LANGUAGE_PORTUGUESE_BRAZIL},
+   {0x2a,   RETRO_LANGUAGE_VIETNAMESE},
+   {0x19,   RETRO_LANGUAGE_RUSSIAN},
+   {0x16,   RETRO_LANGUAGE_PORTUGUESE_PORTUGAL},
+   {0x15,   RETRO_LANGUAGE_POLISH},
+   {0x13,   RETRO_LANGUAGE_DUTCH},
+   {0x12,   RETRO_LANGUAGE_KOREAN},
+   {0x11,   RETRO_LANGUAGE_JAPANESE},
+   {0x10,   RETRO_LANGUAGE_ITALIAN},
+   {0xc,    RETRO_LANGUAGE_FRENCH},
+   {0xa,    RETRO_LANGUAGE_SPANISH},
+   {0x9,    RETRO_LANGUAGE_ENGLISH},
+   {0x8,    RETRO_LANGUAGE_GREEK},
+   {0x7,    RETRO_LANGUAGE_GERMAN},
+   {0x4,    RETRO_LANGUAGE_CHINESE_SIMPLIFIED},  /* neutral */
+   {0x1,    RETRO_LANGUAGE_ARABIC},
    /* MS does not support Esperanto */
    /*{0x0, RETRO_LANGUAGE_ESPERANTO},*/
 };
 
 unsigned short win32_get_langid_from_retro_lang(enum retro_language lang)
 {
-   unsigned i;
+   size_t i;
 
-   for (i = 0; i < sizeof(win32_lang_pairs) / sizeof(win32_lang_pairs[0]); i++)
+   for (i = 0; i < ARRAY_SIZE(win32_lang_pairs); i++)
    {
       if (win32_lang_pairs[i].lang == lang)
          return win32_lang_pairs[i].lang_ident;
@@ -161,9 +159,9 @@ unsigned short win32_get_langid_from_retro_lang(enum retro_language lang)
 
 enum retro_language win32_get_retro_lang_from_langid(unsigned short langid)
 {
-   unsigned i;
+   size_t i;
 
-   for (i = 0; i < sizeof(win32_lang_pairs) / sizeof(win32_lang_pairs[0]); i++)
+   for (i = 0; i < ARRAY_SIZE(win32_lang_pairs); i++)
    {
       if (win32_lang_pairs[i].lang_ident > 0x3ff)
       {
@@ -233,9 +231,6 @@ static bool gfx_init_dwm(void)
       (HRESULT(WINAPI*)(BOOL))dylib_proc(dwm_lib, "DwmEnableMMCSS");
 #else
    DragAcceptFiles_func = DragAcceptFiles;
-#if 0
-   mmcss                = DwmEnableMMCSS;
-#endif
 #endif
 
    if (mmcss)
@@ -266,20 +261,21 @@ static void gfx_set_dwm(void)
 
    if (!composition_enable)
    {
-      RARCH_ERR("Did not find DwmEnableComposition ...\n");
+      RARCH_ERR("Did not find DwmEnableComposition.\n");
       return;
    }
 
    ret = composition_enable(!disable_composition);
    if (FAILED(ret))
-      RARCH_ERR("Failed to set composition state ...\n");
+      RARCH_ERR("Failed to set composition state.\n");
    if (disable_composition)
       g_plat_win32_flags |= PLAT_WIN32_FLAG_DWM_COMPOSITION_DISABLED;
 }
 
-static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
+static size_t frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
 {
-   char buildStr[11]      = {0};
+   size_t _len            = 0;
+   char build_str[11]     = {0};
    bool server            = false;
    const char *arch       = "";
 
@@ -287,15 +283,24 @@ static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
    /* Windows 2000 and later */
    SYSTEM_INFO si         = {{0}};
    OSVERSIONINFOEX vi     = {0};
+#ifndef _MSC_VER
+   /* Vista and later, MSYS2/MINGW64 build */
+   const char win_ver_reg_key[]    = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion";
+   const DWORD reg_read_flags      = RRF_RT_REG_SZ; /* Only read strings (REG_SZ) */
+   const int ProductName_2nd_digit = 9; /* second digit in the string 'Windows 10' */
+   char str_ProductName[64]        = {0};
+   char str_DisplayVersion[64]     = {0};
+   char str_LCUVer[64]             = {0};
+   DWORD key_type                  = 0; /* null pointer */
+   DWORD data_size                 = 0;
+   long reg_read_result;
+   bool read_success               = TRUE;
+   /* end Vista and later; still within Windows 2000 and later block */
+#endif
+
    vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
 
    GetSystemInfo(&si);
-
-   /* Available from NT 3.5 and Win95 */
-   GetVersionEx((OSVERSIONINFO*)&vi);
-
-   server = vi.wProductType != VER_NT_WORKSTATION;
-
    switch (si.wProcessorArchitecture)
    {
       case PROCESSOR_ARCHITECTURE_AMD64:
@@ -310,6 +315,74 @@ static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
       default:
          break;
    }
+
+#ifndef _MSC_VER
+   /* Vista and later, MSYS2/MINGW64 build
+    * Check for Win11 by looking for a specific Registry value.
+    * The behavior of GetVersionEx is changed under Win11 and no longer provides
+    * relevant data. If the specific Registry value is present, read version data
+    * directly from registry and skip remainder of function.
+    * Each read is paired for string values; the first gets the size of the
+    * string (read into data_size); the second passes data_size back as an
+    * argument and reads the actual string. */
+   reg_read_result = RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "LCUVer",
+         reg_read_flags, &key_type, 0, &data_size);
+
+   if (reg_read_result == ERROR_SUCCESS)
+   {
+      if (RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "LCUVer",
+            reg_read_flags, &key_type, str_LCUVer, &data_size) != ERROR_SUCCESS)
+         read_success = FALSE;
+
+      if (RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "ProductName",
+            reg_read_flags, &key_type, 0, &data_size) != ERROR_SUCCESS)
+         read_success = FALSE;
+
+      if (RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "ProductName",
+            reg_read_flags, &key_type, str_ProductName, &data_size) != ERROR_SUCCESS)
+         read_success = FALSE;
+
+      if (RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "DisplayVersion",
+            reg_read_flags, &key_type, 0, &data_size) != ERROR_SUCCESS)
+         read_success = FALSE;
+
+      if (RegGetValue(HKEY_LOCAL_MACHINE, win_ver_reg_key, "DisplayVersion",
+            reg_read_flags, &key_type, str_DisplayVersion, &data_size) != ERROR_SUCCESS)
+         read_success = FALSE;
+
+      if (read_success)
+      {
+         str_ProductName[ProductName_2nd_digit] = '1';
+         /* Even the version in the Registry still says Windows 10 and requires
+          * string manipulation. */
+
+          _len = strlcpy(s, str_ProductName, len);
+
+         if (!string_is_empty(arch))
+         {
+            _len += strlcpy(s + _len, " ",  len - _len);
+            _len += strlcpy(s + _len, arch, len - _len);
+         }
+         _len = strlcpy(s + _len, " ", len - _len);
+         _len = strlcpy(s + _len, str_DisplayVersion, len - _len);
+         _len = strlcpy(s + _len, " (", len - _len);
+         _len = strlcpy(s + _len, str_LCUVer, len - _len);
+         _len = strlcpy(s + _len, ")", len - _len);
+
+         *major = 10;
+         *minor = 0;
+
+         return _len;
+      }
+   }
+   /* End registry-check-and-read code; still within 2000-and-later block */
+#endif
+
+   /* GetVersionEx call changed in Win2K and later */
+   GetVersionEx((OSVERSIONINFO*)&vi);
+
+   server = vi.wProductType != VER_NT_WORKSTATION;
+
 #else
    OSVERSIONINFO vi = {0};
    vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
@@ -325,46 +398,46 @@ static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
       *minor = vi.dwMinorVersion;
 
    if (vi.dwMajorVersion == 4 && vi.dwMinorVersion == 0)
-      snprintf(buildStr, sizeof(buildStr), "%lu", (DWORD)(LOWORD(vi.dwBuildNumber))); /* Windows 95 build number is in the low-order word only */
+      snprintf(build_str, sizeof(build_str), "%lu", (DWORD)(LOWORD(vi.dwBuildNumber))); /* Windows 95 build number is in the low-order word only */
    else
-      snprintf(buildStr, sizeof(buildStr), "%lu", vi.dwBuildNumber);
+      snprintf(build_str, sizeof(build_str), "%lu", vi.dwBuildNumber);
 
    switch (vi.dwMajorVersion)
    {
       case 10:
-         if (atoi(buildStr) >= 21996)
-            strlcpy(s, "Windows 11", len);
+         if (atoi(build_str) >= 21996)
+            _len = strlcpy(s, "Windows 11", len);
          else if (server)
-            strlcpy(s, "Windows Server 2016", len);
+            _len = strlcpy(s, "Windows Server 2016", len);
          else
-            strlcpy(s, "Windows 10", len);
+            _len = strlcpy(s, "Windows 10", len);
          break;
       case 6:
          switch (vi.dwMinorVersion)
          {
             case 3:
                if (server)
-                  strlcpy(s, "Windows Server 2012 R2", len);
+                  _len = strlcpy(s, "Windows Server 2012 R2", len);
                else
-                  strlcpy(s, "Windows 8.1", len);
+                  _len = strlcpy(s, "Windows 8.1", len);
                break;
             case 2:
                if (server)
-                  strlcpy(s, "Windows Server 2012", len);
+                  _len = strlcpy(s, "Windows Server 2012", len);
                else
-                  strlcpy(s, "Windows 8", len);
+                  _len = strlcpy(s, "Windows 8", len);
                break;
             case 1:
                if (server)
-                  strlcpy(s, "Windows Server 2008 R2", len);
+                  _len = strlcpy(s, "Windows Server 2008 R2", len);
                else
-                  strlcpy(s, "Windows 7", len);
+                  _len = strlcpy(s, "Windows 7", len);
                break;
             case 0:
                if (server)
-                  strlcpy(s, "Windows Server 2008", len);
+                  _len = strlcpy(s, "Windows Server 2008", len);
                else
-                  strlcpy(s, "Windows Vista", len);
+                  _len = strlcpy(s, "Windows Vista", len);
                break;
             default:
                break;
@@ -376,22 +449,22 @@ static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
             case 2:
                if (server)
                {
-                  strlcpy(s, "Windows Server 2003", len);
+                  _len = strlcpy(s, "Windows Server 2003", len);
                   if (GetSystemMetrics(SM_SERVERR2))
-                     strlcat(s, " R2", len);
+                     _len += strlcpy(s + _len, " R2", len - _len);
                }
                else
                {
                   /* Yes, XP Pro x64 is a higher version number than XP x86 */
                   if (string_is_equal(arch, "x64"))
-                     strlcpy(s, "Windows XP", len);
+                     _len = strlcpy(s, "Windows XP", len);
                }
                break;
             case 1:
-               strlcpy(s, "Windows XP", len);
+               _len = strlcpy(s, "Windows XP", len);
                break;
             case 0:
-               strlcpy(s, "Windows 2000", len);
+               _len = strlcpy(s, "Windows 2000", len);
                break;
          }
          break;
@@ -400,40 +473,40 @@ static void frontend_win32_get_os(char *s, size_t len, int *major, int *minor)
          {
             case 0:
                if (vi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
-                  strlcpy(s, "Windows 95", len);
+                  _len = strlcpy(s, "Windows 95", len);
                else if (vi.dwPlatformId == VER_PLATFORM_WIN32_NT)
-                  strlcpy(s, "Windows NT 4.0", len);
+                  _len = strlcpy(s, "Windows NT 4.0", len);
                else
-                  strlcpy(s, "Unknown", len);
+                  _len = strlcpy(s, "Unknown", len);
                break;
             case 90:
-               strlcpy(s, "Windows ME", len);
+               _len = strlcpy(s, "Windows ME", len);
                break;
             case 10:
-               strlcpy(s, "Windows 98", len);
+               _len = strlcpy(s, "Windows 98", len);
                break;
          }
          break;
       default:
-         snprintf(s, len, "Windows %i.%i", *major, *minor);
+         _len = snprintf(s, len, "Windows %i.%i", *major, *minor);
          break;
    }
 
    if (!string_is_empty(arch))
    {
-      strlcat(s, " ", len);
-      strlcat(s, arch, len);
+      _len += strlcpy(s + _len, " ",  len - _len);
+      _len += strlcpy(s + _len, arch, len - _len);
    }
 
-   strlcat(s, " Build ", len);
-   strlcat(s, buildStr, len);
+   _len += strlcpy(s + _len, " Build ", len - _len);
+   _len += strlcpy(s + _len, build_str, len - _len);
 
    if (!string_is_empty(vi.szCSDVersion))
    {
-      strlcat(s, " ", len);
-      strlcat(s, vi.szCSDVersion, len);
+      _len += strlcpy(s + _len, " ", len - _len);
+      strlcpy(s + _len, vi.szCSDVersion, len - _len);
    }
-
+   return _len;
 }
 
 static void frontend_win32_init(void *data)
@@ -463,7 +536,7 @@ static void frontend_win32_init(void *data)
 static void init_nvda(void)
 {
 #ifdef HAVE_DYLIB
-   if (     (g_plat_win32_flags & PLAT_WIN32_FLAG_USE_NVDA) 
+   if (     (g_plat_win32_flags & PLAT_WIN32_FLAG_USE_NVDA)
          && !nvda_lib)
    {
       if ((nvda_lib = dylib_load("nvdaControllerClient64.dll")))
@@ -476,8 +549,14 @@ static void init_nvda(void)
       }
    }
 #endif
-   g_plat_win32_flags &= ~PLAT_WIN32_FLAG_USE_NVDA;
-   g_plat_win32_flags |=  PLAT_WIN32_FLAG_USE_POWERSHELL;
+   /* The above code is executed on each accessibility speak event, so
+    * we should only revert to powershell if nvda_lib wasn't loaded previously,
+    * and we weren't able to load it on this call, or we don't HAVE_DYLIB */
+   if ((g_plat_win32_flags & PLAT_WIN32_FLAG_USE_NVDA) && !nvda_lib)
+   {
+      g_plat_win32_flags &= ~PLAT_WIN32_FLAG_USE_NVDA;
+      g_plat_win32_flags |=  PLAT_WIN32_FLAG_USE_POWERSHELL;
+   }
 }
 #endif
 
@@ -569,24 +648,51 @@ static void frontend_win32_env_get(int *argc, char *argv[],
 {
    const char *tmp_dir = getenv("TMP");
    const char *libretro_directory = getenv("LIBRETRO_DIRECTORY");
+   const char *libretro_assets_directory = getenv("LIBRETRO_ASSETS_DIRECTORY");
+   const char* libretro_autoconfig_directory = getenv("LIBRETRO_AUTOCONFIG_DIRECTORY");
+   const char* libretro_cheats_directory = getenv("LIBRETRO_CHEATS_DIRECTORY");
+   const char* libretro_database_directory = getenv("LIBRETRO_DATABASE_DIRECTORY");
+   const char* libretro_system_directory = getenv("LIBRETRO_SYSTEM_DIRECTORY");
+   const char* libretro_video_filter_directory = getenv("LIBRETRO_VIDEO_FILTER_DIRECTORY");
+   const char* libretro_video_shader_directory = getenv("LIBRETRO_VIDEO_SHADER_DIRECTORY");
    if (!string_is_empty(tmp_dir))
       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CACHE],
          tmp_dir, sizeof(g_defaults.dirs[DEFAULT_DIR_CACHE]));
 
    gfx_set_dwm();
 
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_ASSETS],
-      ":\\assets", sizeof(g_defaults.dirs[DEFAULT_DIR_ASSETS]));
+   if (!string_is_empty(libretro_assets_directory))
+      strlcpy(g_defaults.dirs[DEFAULT_DIR_ASSETS], libretro_assets_directory,
+	      sizeof(g_defaults.dirs[DEFAULT_DIR_ASSETS]));
+   else
+       fill_pathname_expand_special(
+	   g_defaults.dirs[DEFAULT_DIR_ASSETS],
+	   ":\\assets", sizeof(g_defaults.dirs[DEFAULT_DIR_ASSETS]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_AUDIO_FILTER],
       ":\\filters\\audio", sizeof(g_defaults.dirs[DEFAULT_DIR_AUDIO_FILTER]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER],
-      ":\\filters\\video", sizeof(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CHEATS],
-      ":\\cheats", sizeof(g_defaults.dirs[DEFAULT_DIR_CHEATS]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_DATABASE],
-      ":\\database\\rdb", sizeof(g_defaults.dirs[DEFAULT_DIR_DATABASE]));
+   if (!string_is_empty(libretro_video_filter_directory))
+       strlcpy(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER],
+	       libretro_video_filter_directory,
+	       sizeof(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER],
+           ":\\filters\\video", sizeof(g_defaults.dirs[DEFAULT_DIR_VIDEO_FILTER]));
+   if (!string_is_empty(libretro_cheats_directory))
+       strlcpy(g_defaults.dirs[DEFAULT_DIR_CHEATS],
+	       libretro_cheats_directory,
+	       sizeof(g_defaults.dirs[DEFAULT_DIR_CHEATS]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CHEATS],
+           ":\\cheats", sizeof(g_defaults.dirs[DEFAULT_DIR_CHEATS]));
+   if (!string_is_empty(libretro_database_directory))
+       strlcpy(g_defaults.dirs[DEFAULT_DIR_DATABASE],
+	       libretro_database_directory,
+	       sizeof(g_defaults.dirs[DEFAULT_DIR_DATABASE]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_DATABASE],
+           ":\\database\\rdb", sizeof(g_defaults.dirs[DEFAULT_DIR_DATABASE]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_PLAYLIST],
-      ":\\playlists", sizeof(g_defaults.dirs[DEFAULT_DIR_ASSETS]));
+      ":\\playlists", sizeof(g_defaults.dirs[DEFAULT_DIR_PLAYLIST]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_RECORD_CONFIG],
       ":\\config\\record", sizeof(g_defaults.dirs[DEFAULT_DIR_RECORD_CONFIG]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_RECORD_OUTPUT],
@@ -601,22 +707,34 @@ static void frontend_win32_env_get(int *argc, char *argv[],
       ":\\thumbnails", sizeof(g_defaults.dirs[DEFAULT_DIR_THUMBNAILS]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_OVERLAY],
       ":\\overlays", sizeof(g_defaults.dirs[DEFAULT_DIR_OVERLAY]));
-#ifdef HAVE_VIDEO_LAYOUT
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_VIDEO_LAYOUT],
-      ":\\layouts", sizeof(g_defaults.dirs[DEFAULT_DIR_VIDEO_LAYOUT]));
-#endif
+   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_OSK_OVERLAY],
+      ":\\overlays\\keyboards", sizeof(g_defaults.dirs[DEFAULT_DIR_OSK_OVERLAY]));
    if (!string_is_empty(libretro_directory))
       strlcpy(g_defaults.dirs[DEFAULT_DIR_CORE], libretro_directory,
             sizeof(g_defaults.dirs[DEFAULT_DIR_CORE]));
    else
       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CORE],
             ":\\cores", sizeof(g_defaults.dirs[DEFAULT_DIR_CORE]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CORE_INFO],
-      ":\\info", sizeof(g_defaults.dirs[DEFAULT_DIR_CORE_INFO]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG],
-      ":\\autoconfig", sizeof(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SHADER],
-      ":\\shaders", sizeof(g_defaults.dirs[DEFAULT_DIR_SHADER]));
+   if (!string_is_empty(libretro_directory))
+      strlcpy(g_defaults.dirs[DEFAULT_DIR_CORE_INFO], libretro_directory,
+            sizeof(g_defaults.dirs[DEFAULT_DIR_CORE_INFO]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CORE_INFO],
+           ":\\info", sizeof(g_defaults.dirs[DEFAULT_DIR_CORE_INFO]));
+   if (!string_is_empty(libretro_autoconfig_directory))
+      strlcpy(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG],
+	      libretro_autoconfig_directory,
+	      sizeof(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG],
+             ":\\autoconfig", sizeof(g_defaults.dirs[DEFAULT_DIR_AUTOCONFIG]));
+   if (!string_is_empty(libretro_video_filter_directory))
+      strlcpy(g_defaults.dirs[DEFAULT_DIR_SHADER],
+	      libretro_video_shader_directory,
+	      sizeof(g_defaults.dirs[DEFAULT_DIR_SHADER]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SHADER],
+             ":\\shaders", sizeof(g_defaults.dirs[DEFAULT_DIR_SHADER]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_CORE_ASSETS],
       ":\\downloads", sizeof(g_defaults.dirs[DEFAULT_DIR_CORE_ASSETS]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SCREENSHOT],
@@ -625,8 +743,13 @@ static void frontend_win32_env_get(int *argc, char *argv[],
       ":\\saves", sizeof(g_defaults.dirs[DEFAULT_DIR_SRAM]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SAVESTATE],
       ":\\states", sizeof(g_defaults.dirs[DEFAULT_DIR_SAVESTATE]));
-   fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SYSTEM],
-      ":\\system", sizeof(g_defaults.dirs[DEFAULT_DIR_SYSTEM]));
+   if (!string_is_empty(libretro_system_directory))
+       strlcpy(g_defaults.dirs[DEFAULT_DIR_SYSTEM],
+	       libretro_system_directory,
+	       sizeof(g_defaults.dirs[DEFAULT_DIR_SYSTEM]));
+   else
+       fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_SYSTEM],
+             ":\\system", sizeof(g_defaults.dirs[DEFAULT_DIR_SYSTEM]));
    fill_pathname_expand_special(g_defaults.dirs[DEFAULT_DIR_LOGS],
       ":\\logs", sizeof(g_defaults.dirs[DEFAULT_DIR_LOGS]));
 
@@ -761,19 +884,16 @@ static void frontend_win32_respawn(char *s, size_t len, char *args)
    if (win32_fork_mode != FRONTEND_FORK_RESTART)
       return;
 
-   fill_pathname_application_path(executable_path,
-         sizeof(executable_path));
+   GetModuleFileName(NULL, executable_path, PATH_MAX_LENGTH);
    path_set(RARCH_PATH_CORE, executable_path);
 
    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);
    memset(&pi, 0, sizeof(pi));
 
-   if (!CreateProcess( executable_path, args,
-      NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
-   {
-      RARCH_ERR("Failed to restart RetroArch\n");
-   }
+   if (!CreateProcess(executable_path, GetCommandLine(),
+         NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
+      RARCH_ERR("Failed to restart RetroArch.\n");
 }
 
 static bool frontend_win32_set_fork(enum frontend_fork fork_mode)
@@ -797,71 +917,6 @@ static bool frontend_win32_set_fork(enum frontend_fork fork_mode)
 #endif
 
 #if defined(_WIN32) && !defined(_XBOX)
-static const char *accessibility_win_language_id(const char* language)
-{
-   if (string_is_equal(language,"en"))
-      return "409";
-   else if (string_is_equal(language,"it"))
-      return "410";
-   else if (string_is_equal(language,"sv"))
-      return "041d";
-   else if (string_is_equal(language,"fr"))
-      return "040c";
-   else if (string_is_equal(language,"de"))
-      return "407";
-   else if (string_is_equal(language,"he"))
-      return "040d";
-   else if (string_is_equal(language,"id"))
-      return "421";
-   else if (string_is_equal(language,"es"))
-      return "040a";
-   else if (string_is_equal(language,"nl"))
-      return "413";
-   else if (string_is_equal(language,"ro"))
-      return "418";
-   else if (string_is_equal(language,"pt_pt"))
-      return "816";
-   else if (string_is_equal(language,"pt_bt") || string_is_equal(language,"pt"))
-      return "416";
-   else if (string_is_equal(language,"th"))
-      return "041e";
-   else if (string_is_equal(language,"ja"))
-      return "411";
-   else if (string_is_equal(language,"sk"))
-      return "041b";
-   else if (string_is_equal(language,"hi"))
-      return "439";
-   else if (string_is_equal(language,"ar"))
-      return "401";
-   else if (string_is_equal(language,"hu"))
-      return "040e";
-   else if (string_is_equal(language,"zh_tw") || string_is_equal(language,"zh"))
-      return "804";
-   else if (string_is_equal(language,"el"))
-      return "408";
-   else if (string_is_equal(language,"ru"))
-      return "419";
-   else if (string_is_equal(language,"nb"))
-      return "414";
-   else if (string_is_equal(language,"da"))
-      return "406";
-   else if (string_is_equal(language,"fi"))
-      return "040b";
-   else if (string_is_equal(language,"zh_hk"))
-      return "0c04";
-   else if (string_is_equal(language,"zh_cn"))
-      return "804";
-   else if (string_is_equal(language,"tr"))
-      return "041f";
-   else if (string_is_equal(language,"ko"))
-      return "412";
-   else if (string_is_equal(language,"pl"))
-      return "415";
-   else if (string_is_equal(language,"cs")) 
-      return "405";
-   return "";
-}
-
 static const char *accessibility_win_language_code(const char* language)
 {
    if (string_is_equal(language,"en"))
@@ -900,13 +955,15 @@ static const char *accessibility_win_language_code(const char* language)
       return "Microsoft Naayf Desktop";
    else if (string_is_equal(language,"hu"))
       return "Microsoft Szabolcs Desktop";
-   else if (string_is_equal(language,"zh_tw") || string_is_equal(language,"zh"))
+   else if (string_is_equal(language, "zh_tw")
+            || string_is_equal(language,"zh-TW")
+            || string_is_equal(language,"zh"))
       return "Microsoft Zhiwei Desktop";
    else if (string_is_equal(language,"el"))
       return "Microsoft Stefanos Desktop";
    else if (string_is_equal(language,"ru"))
       return "Microsoft Pavel Desktop";
-   else if (string_is_equal(language,"nb"))
+   else if (string_is_equal(language,"no") || string_is_equal(language,"nb"))
       return "Microsoft Jon Desktop";
    else if (string_is_equal(language,"da"))
       return "Microsoft Helle Desktop";
@@ -914,7 +971,7 @@ static const char *accessibility_win_language_code(const char* language)
       return "Microsoft Heidi Desktop";
    else if (string_is_equal(language,"zh_hk"))
       return "Microsoft Danny Desktop";
-   else if (string_is_equal(language,"zh_cn"))
+   else if (string_is_equal(language,"zh_cn") || string_is_equal(language,"zh-CN"))
       return "Microsoft Kangkang Desktop";
    else if (string_is_equal(language,"tr"))
       return "Microsoft Tolga Desktop";
@@ -922,17 +979,32 @@ static const char *accessibility_win_language_code(const char* language)
       return "Microsoft Heami Desktop";
    else if (string_is_equal(language,"pl"))
       return "Microsoft Adam Desktop";
-   else if (string_is_equal(language,"cs")) 
+   else if (string_is_equal(language,"cs"))
       return "Microsoft Jakub Desktop";
+   else if (string_is_equal(language,"vi"))
+      return "Microsoft An Desktop";
+   else if (string_is_equal(language,"hr"))
+      return "Microsoft Matej Desktop";
+   else if (string_is_equal(language,"bg"))
+      return "Microsoft Ivan Desktop";
+   else if (string_is_equal(language,"ms"))
+      return "Microsoft Rizwan Desktop";
+   else if (string_is_equal(language,"sl"))
+      return "Microsoft Lado Desktop";
+   else if (string_is_equal(language,"ta"))
+      return "Microsoft Valluvar Desktop";
+   else if (string_is_equal(language,"en_gb"))
+      return "Microsoft George Desktop";
+   else if (string_is_equal(language,"ca") || string_is_equal(language,"ca_ES@valencia"))
+      return "Microsoft Herena Desktop";
    return "";
 }
 
-static bool terminate_win32_process(PROCESS_INFORMATION pi)
+static void terminate_win32_process(PROCESS_INFORMATION pi)
 {
    TerminateProcess(pi.hProcess,0);
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
-   return true;
 }
 
 static PROCESS_INFORMATION g_pi;
@@ -953,12 +1025,12 @@ static bool create_win32_process(char* cmd, const char * input)
       size_t input_len = strlen(input);
       if (!CreatePipe(&rd, &wr, NULL, input_len))
          return false;
-      
+
       SetHandleInformation(rd, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
-      
+
       WriteFile(wr, input, input_len, &dummy, NULL);
       CloseHandle(wr);
-      
+
       si.dwFlags    |= STARTF_USESTDHANDLES;
       si.hStdInput   = rd;
       si.hStdOutput  = GetStdHandle(STD_OUTPUT_HANDLE);
@@ -993,12 +1065,12 @@ static bool is_narrator_running_windows(void)
    {
       long res = nvdaController_testIfRunning_func();
 
-      if (res != 0) 
+      if (res != 0)
       {
          /* The running nvda service wasn't found, so revert
             back to the powershell method
          */
-         RARCH_ERR("Error communicating with NVDA\n");
+         RARCH_ERR("Error communicating with NVDA.\n");
          g_plat_win32_flags |=  PLAT_WIN32_FLAG_USE_POWERSHELL;
          g_plat_win32_flags &= ~PLAT_WIN32_FLAG_USE_NVDA;
          return false;
@@ -1009,11 +1081,11 @@ static bool is_narrator_running_windows(void)
 #ifdef HAVE_SAPI
    else
    {
-      SPVOICESTATUS pStatus;
-      if (pVoice)
+      if (voice_ptr)
       {
-         ISpVoice_GetStatus(pVoice, &pStatus, NULL);
-         if (pStatus.dwRunningState == SPRS_IS_SPEAKING)
+         SPVOICESTATUS status_ptr;
+         ISpVoice_GetStatus(voice_ptr, &status_ptr, NULL);
+         if (status_ptr.dwRunningState == SPRS_IS_SPEAKING)
             return true;
       }
    }
@@ -1027,10 +1099,7 @@ static bool accessibility_speak_windows(int speed,
    char cmd[512];
    const char *voice      = get_user_language_iso639_1(true);
    const char *language   = accessibility_win_language_code(voice);
-   const char *langid     = accessibility_win_language_id(voice);
-   bool res               = false;
-   const char* speeds[10] = {"-10", "-7.5", "-5", "-2.5", "0", "2", "4", "6", "8", "10"};
-   size_t nbytes_cmd      = 0;
+   const char *speeds[10] = {"-10", "-7.5", "-5", "-2.5", "0", "2", "4", "6", "8", "10"};
    if (speed < 1)
       speed               = 1;
    else if (speed > 10)
@@ -1040,16 +1109,15 @@ static bool accessibility_speak_windows(int speed,
    {
       if (is_narrator_running_windows())
          return true;
-   
    }
 #ifdef HAVE_NVDA
    init_nvda();
 #endif
-   
+
    if (g_plat_win32_flags & PLAT_WIN32_FLAG_USE_POWERSHELL)
    {
-      const char * template_lang = "powershell.exe -NoProfile -WindowStyle Hidden -Command \"Add-Type -AssemblyName System.Speech; $synth = New-Object System.Speech.Synthesis.SpeechSynthesizer; $synth.SelectVoice(\\\"%s\\\"); $synth.Rate = %s; $synth.Speak($input);\"";
-      const char * template_nolang = "powershell.exe -NoProfile -WindowStyle Hidden -Command \"Add-Type -AssemblyName System.Speech; $synth = New-Object System.Speech.Synthesis.SpeechSynthesizer; $synth.Rate = %s; $synth.Speak($input);\"";
+      const char *template_lang = "powershell.exe -NoProfile -WindowStyle Hidden -Command \"Add-Type -AssemblyName System.Speech; $synth = New-Object System.Speech.Synthesis.SpeechSynthesizer; $synth.SelectVoice(\\\"%s\\\"); $synth.Rate = %s; $synth.Speak($input);\"";
+      const char *template_nolang = "powershell.exe -NoProfile -WindowStyle Hidden -Command \"Add-Type -AssemblyName System.Speech; $synth = New-Object System.Speech.Synthesis.SpeechSynthesizer; $synth.Rate = %s; $synth.Speak($input);\"";
       if (language && language[0] != '\0')
          snprintf(cmd, sizeof(cmd), template_lang, language, speeds[speed-1]);
       else
@@ -1067,12 +1135,15 @@ static bool accessibility_speak_windows(int speed,
       wchar_t        *wc = utf8_to_utf16_string_alloc(speak_text);
       long res           = nvdaController_testIfRunning_func();
 
-      if (!wc || res != 0) 
+      if (!wc || res != 0)
       {
-         RARCH_ERR("Error communicating with NVDA\n");
+         RARCH_ERR("Error communicating with NVDA.\n");
+         /* Fallback on powershell immediately and retry */
+         g_plat_win32_flags &= ~PLAT_WIN32_FLAG_USE_NVDA;
+         g_plat_win32_flags |= PLAT_WIN32_FLAG_USE_POWERSHELL;
          if (wc)
             free(wc);
-         return false;
+         return accessibility_speak_windows(speed, speak_text, priority);
       }
 
       nvdaController_cancelSpeech_func();
@@ -1089,26 +1160,26 @@ static bool accessibility_speak_windows(int speed,
    {
       HRESULT hr;
       /* stop the old voice if running */
-      if (pVoice)
+      if (voice_ptr)
       {
          CoUninitialize();
-         ISpVoice_Release(pVoice);
+         ISpVoice_Release(voice_ptr);
       }
-      pVoice = NULL;
+      voice_ptr = NULL;
 
       /* Play the new voice */
       if (FAILED(CoInitialize(NULL)))
          return NULL;
 
       hr = CoCreateInstance(&CLSID_SpVoice, NULL,
-            CLSCTX_ALL, &IID_ISpVoice, (void **)&pVoice);
+            CLSCTX_ALL, &IID_ISpVoice, (void **)&voice_ptr);
 
       if (SUCCEEDED(hr))
       {
-         wchar_t        *wc = utf8_to_utf16_string_alloc(speak_text);
+         wchar_t *wc = utf8_to_utf16_string_alloc(speak_text);
          if (!wc)
             return false;
-         hr = ISpVoice_Speak(pVoice, wc, SPF_ASYNC /*SVSFlagsAsync*/, NULL);
+         hr = ISpVoice_Speak(voice_ptr, wc, SPF_ASYNC /*SVSFlagsAsync*/, NULL);
          free(wc);
       }
    }
